@@ -3,8 +3,8 @@ import { Parser, Handler, ParserOptions } from "./Parser.js";
  * NOTE: If either of these two imports produces a type error,
  * please update your @types/node dependency!
  */
-import { Writable } from "stream";
-import { StringDecoder } from "string_decoder";
+import { Writable } from "node:stream";
+import { StringDecoder } from "node:string_decoder";
 
 // Following the example in https://nodejs.org/api/stream.html#stream_decoding_buffers_in_a_writable_stream
 function isBuffer(_chunk: string | Buffer, encoding: string): _chunk is Buffer {
@@ -28,16 +28,16 @@ export class WritableStream extends Writable {
     override _write(
         chunk: string | Buffer,
         encoding: string,
-        cb: () => void
+        callback: () => void
     ): void {
         this._parser.write(
             isBuffer(chunk, encoding) ? this._decoder.write(chunk) : chunk
         );
-        cb();
+        callback();
     }
 
-    override _final(cb: () => void): void {
+    override _final(callback: () => void): void {
         this._parser.end(this._decoder.end());
-        cb();
+        callback();
     }
 }
