@@ -1,5 +1,5 @@
 import { Parser, ParserOptions } from "./Parser.js";
-export { Parser, type ParserOptions };
+export { Parser, type ParserOptions } from "./Parser.js";
 
 import {
     DomHandler,
@@ -9,9 +9,14 @@ import {
     Document,
 } from "domhandler";
 
-export { DomHandler, type DomHandlerOptions };
+export {
+    DomHandler,
+    // Old name for DomHandler
+    DomHandler as DefaultHandler,
+    type DomHandlerOptions,
+} from "domhandler";
 
-type Options = ParserOptions & DomHandlerOptions;
+export type Options = ParserOptions & DomHandlerOptions;
 
 // Helper methods
 
@@ -19,7 +24,7 @@ type Options = ParserOptions & DomHandlerOptions;
  * Parses the data, returns the resulting document.
  *
  * @param data The data that should be parsed.
- * @param options Optional options for the parser and DOM builder.
+ * @param options Optional options for the parser and DOM handler.
  */
 export function parseDocument(data: string, options?: Options): Document {
     const handler = new DomHandler(undefined, options);
@@ -33,7 +38,7 @@ export function parseDocument(data: string, options?: Options): Document {
  * Use `parseDocument` to get the `Document` node instead.
  *
  * @param data The data that should be parsed.
- * @param options Optional options for the parser and DOM builder.
+ * @param options Optional options for the parser and DOM handler.
  * @deprecated Use `parseDocument` instead.
  */
 export function parseDOM(data: string, options?: Options): ChildNode[] {
@@ -42,16 +47,36 @@ export function parseDOM(data: string, options?: Options): ChildNode[] {
 /**
  * Creates a parser instance, with an attached DOM handler.
  *
- * @param cb A callback that will be called once parsing has been completed.
- * @param options Optional options for the parser and DOM builder.
- * @param elementCb An optional callback that will be called every time a tag has been completed inside of the DOM.
+ * @param callback A callback that will be called once parsing has been completed, with the resulting document.
+ * @param options Optional options for the parser and DOM handler.
+ * @param elementCallback An optional callback that will be called every time a tag has been completed inside of the DOM.
+ */
+export function createDocumentStream(
+    callback: (error: Error | null, document: Document) => void,
+    options?: Options,
+    elementCallback?: (element: Element) => void
+): Parser {
+    const handler: DomHandler = new DomHandler(
+        (error: Error | null) => callback(error, handler.root),
+        options,
+        elementCallback
+    );
+    return new Parser(handler, options);
+}
+/**
+ * Creates a parser instance, with an attached DOM handler.
+ *
+ * @param callback A callback that will be called once parsing has been completed, with an array of root nodes.
+ * @param options Optional options for the parser and DOM handler.
+ * @param elementCallback An optional callback that will be called every time a tag has been completed inside of the DOM.
+ * @deprecated Use `createDocumentStream` instead.
  */
 export function createDomStream(
-    cb: (error: Error | null, dom: ChildNode[]) => void,
+    callback: (error: Error | null, dom: ChildNode[]) => void,
     options?: Options,
-    elementCb?: (element: Element) => void
+    elementCallback?: (element: Element) => void
 ): Parser {
-    const handler = new DomHandler(cb, options, elementCb);
+    const handler = new DomHandler(callback, options, elementCallback);
     return new Parser(handler, options);
 }
 
@@ -64,12 +89,13 @@ export {
  * All of the following exports exist for backwards-compatibility.
  * They should probably be removed eventually.
  */
-import * as ElementType from "domelementtype";
-export { ElementType };
+export * as ElementType from "domelementtype";
 
-import { getFeed, Feed } from "domutils";
+import { getFeed, type Feed } from "domutils";
 
-export { getFeed };
+export { getFeed, type Feed } from "domutils";
+
+const parseFeedDefaultOptions = { xmlMode: true };
 
 /**
  * Parse a feed.
@@ -79,12 +105,9 @@ export { getFeed };
  */
 export function parseFeed(
     feed: string,
-    options: ParserOptions & DomHandlerOptions = { xmlMode: true }
+    options: Options = parseFeedDefaultOptions
 ): Feed | null {
     return getFeed(parseDOM(feed, options));
 }
 
 export * as DomUtils from "domutils";
-
-// Old name for DomHandler
-export { DomHandler as DefaultHandler };
